*! lr*() version 2.0.0

version 15
set matastrict on

mata:

struct lrinfo 
{
        // --------------------------------------- inputs
        pointer(real colvector) scalar  y     
        pointer(real matrix) scalar     X     
        real scalar                     cons  

        // --------------------------------------- derived

        real matrix     XX              // X'X; K x K 
        real colvector  Xy              // X'y; K x K
        real matrix     XXinv           // (X'X)^(-1);   K x K
        real colvector  b               // coefficients; K x 1
        real matrix     V               // variance matrix (VCE); K x K
        real scalar     s2              // variance of e
        real scalar     yy              // y'y
        real scalar     ee              // e'e, error sum of squares
        real scalar     K_adj
        real scalar     k_adj

        /*
            Deviation from mean notation:
                    S = X - mean(X)    X as deviation from mean
                    t = y - mean(y)    y as deviation from mean
        */

        real matrix    SS                // S'S; k x k 
        real colvector St                // S't; K x 1 (sic)
        real matrix    SSinv             // (S'S)^(-1); K x K (sic) 
        real rowvector mean_X            // means of rows of X; 1 x k
        real scalar    mean_y            // mean  of row  of y; 1 x 1
}

void lrinit(struct lrinfo scalar r)
{
        r.XX      = .z
        r.Xy      = .z
        r.XXinv   = .z 
        r.b       = .z
        r.V       = .z
        r.s2      = .z
        r.yy      = .z
        r.ee      = .z

        r.SS      = .z
        r.St      = .z
        r.mean_X  = .z
        r.mean_y  = .z
        r.SSinv   = .z
        r.K_adj   = .z 
        r.k_adj   = .z 
}


/* -------------------------------------------------- entry points --- */

struct lrinfo scalar lrset( real colvector y, 
                            real matrix    X,  
                            real scalar    cons )
{
        struct lrinfo scalar   r

        assert(rows(y)==rows(X))

        r.y    = &y
        r.X    = &X
        r.cons = (cons!=0)

        lrinit(r)              // initialize derived vars

        return(r) 
}

real scalar lr_N(struct lrinfo scalar r) return(rows(*r.X))
real scalar lr_k(struct lrinfo scalar r) return(cols(*r.X))
real scalar lr_K(struct lrinfo scalar r) return(cols(*r.X) + r.cons)


/*
        Throughout this code, distinguish between k and K:

                k = # of independent variables EXCLUDING intercept
                K = # of independent variables INCLUDING intercept

        In models without an intercept, K = k
        In models with    an intercept, K = k + 1

        Note: *r.X is k x k, X does not include vector of 1s, and 
                 b is K x 1, coefficient include intercept if there is one.

        To code the equivalent of X*b, you extract the first k elements
        of b and add the intercept separately.  To make this easier
        think of

                 b = (b_X, b_c)   

        lr_b_X() returns b_X:  1 x k
        lr_b_c() returns b_c:  1 x 1  containing b[K] or 0. 

        Thus, X*b can be calculated as (*r.X)*lr_b_X(r) :+ lr_b_c(r) 
        in all cases.
*/


real colvector lr_b_X(struct lrinfo scalar r)
{
        if (r.cons) {         
                if (lr_k(r)) return( lr_b(r)[| 1\lr_k(r) |])
                else         return( J(0, 1, .))
        }
        else                 return(lr_b(r))
}


real scalar lr_b_c(struct lrinfo scalar r) 
{
        return(r.cons ? lr_b(r)[lr_K(r)] : 0) 
}


real colvector lr_b(struct lrinfo scalar r)
{
        if (r.b == .z) {
                if (r.cons) {
                        r.b          = lr_SSinv(r) * lr_St(r)
                        r.b[lr_K(r)] = lr_mean_y(r) - lr_mean_X(r)*lr_b_X(r)

                }
                else    r.b = lr_XXinv(r) * lr_Xy(r)
        }
        return(r.b)
}


real matrix lr_V(struct lrinfo scalar r)
{
        if (r.V == .z) {
                r.V = lr_s2(r) * (r.cons ? lr_SSinv(r) : lr_XXinv(r))
        }
        return(r.V)
}

real scalar lr_s2(struct lrinfo scalar r)
{
        if (r.s2 == .z) {
                r.s2 = lr_ee(r) / ( lr_N(r) - lr_K_adj(r) )
        }
        return(r.s2)
}

real scalar lr_ee(struct lrinfo scalar r)
{
        real colvector   e

        if (r.ee == .z) {
                e = *r.y - lr_yhat(r)
                r.ee = quadcross(e, e)
        }
        return(r.ee)
}

real colvector lr_yhat(struct lrinfo scalar r)
{
        return( (*r.X)*lr_b_X(r) :+ lr_b_c(r) )
}


/* ------------------------------------------internal subroutines --- */

real matrix lr_XX(struct lrinfo scalar r)
{
        if (r.XX == .z) {
                r.XX =  quadcross(*r.X, r.cons,  *r.X, r.cons)
        }
        return(r.XX)
}

real colvector lr_Xy(struct lrinfo scalar r)
{
        if (r.Xy == .z) {
                r.Xy =  quadcross(*r.X, r.cons,  *r.y, 0)
        }
        return(r.Xy)
}

real scalar lr_yy(struct lrinfo scalar r)
{
        if (r.yy == .z) {
                r.yy = quadcross(*r.y, *r.y)
        }
        return(r.yy)
}

real matrix lr_XXinv(struct lrinfo scalar r)
{
        if (r.XXinv == .z) {
                r.XXinv = invsym(lr_XX(r))
        }
        return(r.XXinv)
}

real matrix lr_SSinv(struct lrinfo scalar r)
{
        if (r.SSinv == .z) {
                if (r.cons) calc_SSinv_cons(r)
        }
        return(r.SSinv)
}


void calc_SSinv_cons(struct lrinfo scalar r) 
{
        real scalar        K, k, i
        real rowvector        v
        real matrix        kxk, rowK, colK

        K = lr_K(r)                // # X vars including constant
        k = lr_k(r)                // # x vars excluding constant

        /* 
            Calculate inverse of S'S = lr_SS().
            This routine called only if cons, meaning K = k + 1.
            lr_SS() is k x k -- it excludes constant. 
            We want the full K x K inverse. 
        */

        // ---------------------------------------- SSinv 1 x 1 case ---
        if (K==1) {
                r.SSinv = 1/lr_N(r)
                return
        }

        // ----------------------------------- SSinv 2 x 2 or larger ---

        r.SSinv = J(K, K, .) // matrix now K x K

                             // Range subscripts:
        kxk  = (1,1 \ k,k)   //   k x k submatrix  rows & cols 1..k
        rowK = (K,1 \ K,k)   //   1 x k submatrix; row K
        colK = (1,K \ k,K)   //   k x 1 submatrix; col K

        r.SSinv[|kxk |]     = invsym(lr_SS(r))

        r.SSinv[|rowK|] = v = -(lr_mean_X(r) * r.SSinv[|kxk|])
        r.SSinv[|colK|] = v'

        // r.SSinv[ K,K  ]     = 1/lr_N(r) - lr_mean_X(r)*r.SSinv[|colK|]
        r.SSinv[ K,K  ]     = 1/lr_N(r) - lr_mean_X(r)*v'
}


real matrix lr_SS(struct lrinfo scalar r)
{
        if (r.SS == .z) {
                if (r.cons) {
                        r.SS = quadcrossdev(*r.X, 0, lr_mean_X(r), 
                                            *r.X, 0, lr_mean_X(r) )
                }
        }
        return(r.SS)
}


real colvector lr_St(struct lrinfo scalar r)
{
        if (r.St == .z) {
                r.St = quadcrossdev(*r.X,      0, lr_mean_X(r), 
                                    *r.y,      0, lr_mean_y(r)) \ 0
        }
        return(r.St)
}


real rowvector lr_mean_X(struct lrinfo scalar r) 
{
        if (r.mean_X == .z) r.mean_X = mean(*r.X)
        return(r.mean_X)
}


real scalar lr_mean_y(struct lrinfo scalar r)
{
        if (r.mean_y == .z) r.mean_y = mean(*r.y)
        return(r.mean_y)
}


real scalar lr_K_adj(struct lrinfo scalar r)
{
        if (r.K_adj == .z) {
                if (r.cons) r.K_adj = colsum(diagonal(lr_SSinv(r)) :!= 0)
                else        r.K_adj = colsum(diagonal(lr_XXinv(r)) :!= 0)
        }
        return(r.K_adj)
}


real scalar lr_k_adj(struct lrinfo scalar r)
{
        if (r.k_adj == .z) {
                r.k_adj = lr_k(r) - (lr_K(r)-lr_K_adj(r))
        }
        return(r.k_adj) 
}

end
